/*
**
**  Copyright (C) 1996, OFFIS
**
**  This software and supporting documentation were developed by
**
**    Kuratorium OFFIS e.V.
**    Forschungsbereich 2: Kommunikationssysteme
**    Escherweg 2
**    D-26121 Oldenburg, Germany
**
**  for CEN/TC251/WG4 as a contribution to the Computer Assisted Radiology
**  (CAR) 1996 DICOM Demonstration.
**
**  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
**  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
**  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
**  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
**
**  Copyright of the software  and  supporting  documentation  is,  unless
**  otherwise stated, owned by OFFIS, and free access is hereby granted as
**  a license to  use  this  software,  copy  this  software  and  prepare
**  derivative works based upon this software.  However, any  distribution
**  of this software source code or supporting documentation or derivative
**  works  (source code and  supporting documentation)  must  include  the
**  three paragraphs of this copyright notice.
**
*/

/*
** Modality Worklist Server Central Test Node (ctn) Main Program
**
** Author: 	Andrew Hewett
**		Kuratorium OFFIS e.V., Oldenburg, Germany
** Created:	03/1996
**
** Last Update:		hewett
** Update Date:		1996/08/29 16:16:10
** Source File:		/share/dicom/cvs-depot/dcmtk/wlistctn/apps/wlistctn.cc,v
** CVS/RCS Revision:	1.4
** Status:		Exp
**
** CVS/RCS Log at end of file
*/

#include "osconfig.h"    /* make sure OS specific configuration is included first */


#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#include <time.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <errno.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#include "wlistctn.h"
#include "dimse.h"
#include "sce.h"
#ifndef NODISPLAY
#include "dispuser.h"
#endif
#include "wrklstdb.h"
#include "dcdebug.h"

#define APPLICATIONTITLE	"WLISTCTN"

typedef struct _ctn_processslot CTN_ProcessSlot;

struct _ctn_processslot {
	DIC_NODENAME		peerName;
	DIC_AE			callingAETitle;
	DIC_AE			calledAETitle;
	int			processId;
	time_t			startTime;
	BOOLEAN			hasStorageAbility;
	CTN_ProcessSlot		*next;
};

typedef struct {
	int			pcnt;	
	CTN_ProcessSlot		*plist;
} CTN_ProcessTable;

static CTN_ProcessTable	processTable;

char *supportedAbstractSyntaxes[] = {
    UID_VerificationSOPClass, 
    UID_FINDModalityWorklistInformationModel
};

int numberOfSupportedAbstractSyntaxes = DIM_OF(supportedAbstractSyntaxes);

/* Reasons for refusing association */
typedef enum {
	CTN_TooManyAssociations,
	CTN_CannotFork,
	CTN_BadAppContext,
	CTN_BadAEPeer,
	CTN_BadAEService,
	CTN_NoReason
} CTN_RefuseReason;



char* progname = NULL;
BOOLEAN verbose = FALSE;
BOOLEAN debug = FALSE;

char* baseWorklistDBPath = "/home/www/wlist";

static BOOLEAN singleProcess = FALSE;
static BOOLEAN rejectWhenNoImplementationClassUID = FALSE;
static BOOLEAN refuse = FALSE;
static int maxReceivePDULength = ASC_DEFAULTMAXPDU;

static BOOLEAN disableDisplay = TRUE;

static BOOLEAN checkFindIdentifier = FALSE;

static int maxAssociations = 20;	/* default */

static void
shortusage()
{
    fprintf(stderr, "\
usage: %s [-c <db-path>][-D][-v][-d][-s][-r][-b n] port\n", 
	progname);
}

static void
fullusage()
{
    shortusage();
    fprintf(stderr, "\
parameters:\n\
    port	 tcp/ip port number to listen on\n\
options:\n\
    -c <db-path> base path of worklist database\n\
    -D		 enable display\n\
    -v		 verbose mode\n\
    -d		 debug mode\n\
    -s		 single process\n\
    -r		 refuse association\n\
    -b n	 set max receive PDU to n bytes\n"
    );
    exit(1);
}

static void 
usage()
{
    shortusage();
    exit(1);
}

void errmsg(const char* msg, ...)
{
    va_list args;
	
    fprintf(stderr, "%s: ", progname);
    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
    fprintf(stderr, "\n");
}

static CONDITION waitForAssociation(T_ASC_Network *net);
static void cleanChildren();
static CONDITION negotiateAssociation(T_ASC_Association *assoc);

int 
main(int argc, char *argv[])
{
    CONDITION cond;
    T_ASC_Network *net;
    int i;
    int port = 0;

    /* strip any leading path from program name */
    if ((progname = (char *)strrchr(argv[0], PATH_SEPARATOR)) != NULL) {
	progname++;
    } else {
	progname = argv[0];
    }

    if (argc < 2) {
	fullusage();
    }
    /* parse program arguments */
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	switch (argv[i][1]) {
	case 'D':
	    disableDisplay = FALSE;
#ifdef NODISPLAY
	    fprintf(stderr, "Warning: not compiled for ctndisplay\n");
#endif	    
	    break;
	case 'v':
	    verbose = TRUE;
	    break;
	case 'd':
	    debug = TRUE;
	    verbose = TRUE;
	    break;
	case 'c':
	    if (i++ < argc) {
		baseWorklistDBPath = argv[i];
	    } else
		usage();
	    break;
	case 's':
	    singleProcess = TRUE;
	    break;
	case 'r':
	    refuse = TRUE;
	    break;
	case 'X':
	    switch (argv[i][2]) {
	    case 'F':
	        checkFindIdentifier = TRUE;
	        break;
	    default:
	        usage();
	        break;
	    }
	    break;
	case 'b':
	    if (((i + 1) < argc) &&
		(sscanf(argv[i + 1], "%d", &maxReceivePDULength) == 1)) {
		i++;		/* Maximum Receive PDU Length parsed */
		if (maxReceivePDULength < ASC_MINIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too small",
			   maxReceivePDULength);
		    usage();
		} else if (maxReceivePDULength > ASC_MAXIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too big",
			   maxReceivePDULength);
		    usage();
		}
	    } else {
		usage();
	    }
	    break;
	default:
	    fullusage();
	}
    }

#ifdef USE_MALLOC_DEBUG
#ifdef HAVE_MALLOC_DEBUG
#define MALLOCDEBUGLEVEL 16
    if (verbose) {
        printf("malloc_debug on (%d)\n", MALLOCDEBUGLEVEL);
    }
    malloc_debug(MALLOCDEBUGLEVEL);
#endif
#endif

    if (port == 0) {
        /* not set, use default */
	port = 104;
    }

    if (i > (argc-1)) {
	errmsg("too many arguments");
	usage();
    }
    if (i != (argc-1)) {
	errmsg("missing port number");
	usage();
    }

    /* get port number to listen on */
    if (sscanf(argv[i], "%d", &port) != 1) {
	errmsg("bad port number: %s", argv[i]);
	usage();
    }

    if (access(baseWorklistDBPath, R_OK) < 0) {
	errmsg("cannot access base worklist db directory");
	errmsg("%s: %s", baseWorklistDBPath, strerror(errno));
	usage();
    }

    DUL_Debug(debug);
    DIMSE_debug(debug);
    SetDebugLevel(((debug)?(3):(0)));	/* dcmdata debugging */
    

    processTable.pcnt = 0;
    processTable.plist = NULL;

    /* if port is privileged we must be as well */
    if (port < 1024) {
        if (geteuid() != 0) {
	    errmsg("cannot listen on port %d, insufficient privileges", port);
	    usage();
	}
    }
    
    cond = ASC_initializeNetwork(NET_ACCEPTOR, port, 1000, &net);
    if (!SUCCESS(cond)) {
	errmsg("Error initialising network:");
	COND_DumpConditions();
	exit(1);
    }

    /* return to normal uid so that we can't do too much damage in case
     * things go very wrong.   Only works if the program is setuid root,
     * and run by another user.  Running as root user may be
     * potentially disasterous if this program screws up badly.
     */
    setuid(getuid());

    /* initialise display interface */
#ifndef NODISPLAY
    DSU_setTraceLevel(debug);
    if (disableDisplay) {
        DSU_setDisplayMode(FALSE);
    } else {
        DSU_setDisplayMode(TRUE);
    }
    DSU_initDisplay(DSU_CTNUser);
#endif

    /* loop waiting for associations */
    while (SUCCESS(cond)) {

	cond = waitForAssociation(net);

	if (!singleProcess) {
	    cleanChildren();	/* clean up any child processes */
	}
    }

    cond = ASC_dropNetwork(&net);
    if (!SUCCESS(cond)) {
	errmsg("Error dropping network:");
	COND_DumpConditions();
	exit(1);
    }
    return 0;

}

/*
 * Process Table List Handling
 */

static void
addProcessToTable(int pid, T_ASC_Association * assoc)
{
    CTN_ProcessSlot	*ps;

    if ((ps = (CTN_ProcessSlot*)malloc(sizeof(*ps))) == NULL) {
	errmsg("out of memory");
	return;
    }
    /* fill in info */
    ASC_getPresentationAddresses(assoc->params, ps->peerName, NULL);
    ASC_getAPTitles(assoc->params, ps->callingAETitle, 
	ps->calledAETitle, NULL);
    ps->processId = pid;
    ps->startTime = time(NULL);
    ps->hasStorageAbility = FALSE;
    /* add to start of list */
    processTable.pcnt++;
    ps->next = processTable.plist;
    processTable.plist = ps;
}

#if 0
/* Unused */
static CTN_ProcessSlot* procWithCalledAETitle(const char* aeTitle)
{
    CTN_ProcessSlot	*ps;
    BOOLEAN		found = FALSE;

    if (aeTitle == NULL) return NULL;

    ps = processTable.plist;
    while (ps && !found) {
	found = strcmp(ps->calledAETitle, aeTitle) == 0;
	if (!found) ps = ps->next;
    }
    return ps;
}

static CTN_ProcessSlot* procWithCallingAETitle(const char* aeTitle)
{
    CTN_ProcessSlot	*ps;
    BOOLEAN		found = FALSE;

    if (aeTitle == NULL) return NULL;

    ps = processTable.plist;
    while (ps && !found) {
	found = strcmp(ps->callingAETitle, aeTitle) == 0;
	if (!found) ps = ps->next;
    }
    return ps;
}
#endif

#if 0
/* currently unused */
static int
countProcessesForCalledAE(const char* aeTitle)
{
    int n = 0;
    CTN_ProcessSlot	*ps;

    if (aeTitle == NULL) return 0;

    ps = processTable.plist;
    while (ps) {
	if (strcmp(ps->calledAETitle, aeTitle) == 0) {
	    n++;
	}
	ps = ps->next;
    }
    return n;
}
#endif

#if 0
/* Unused */
static CTN_ProcessSlot* procWithPeerName(const char* peerName)
{
    CTN_ProcessSlot	*ps;
    BOOLEAN		found = FALSE;

    if (peerName == NULL) return NULL;

    ps = processTable.plist;
    while (ps && !found) {
	found = strcmp(ps->peerName, peerName) == 0;
	if (!found) ps = ps->next;
    }
    return ps;
}
#endif

static int
countChildProcesses()
{
    int n = 0;
    CTN_ProcessSlot	*ps;

    ps = processTable.plist;
    while (ps) {
	n++;
	ps = ps->next;
    }
    return n;
}

static void
removeProcessFromTable(int pid)
{
    CTN_ProcessSlot	*ps, *pps;
    BOOLEAN		found = FALSE;

    pps = NULL;
    ps = processTable.plist;

    while (ps && !found) {
	found = (ps->processId == pid);
	if (!found) {
	    pps = ps;
	    ps = ps->next;
	}
    }

    if (!found) {
	errmsg("WARNING: process %d not found", pid);
	return;
    }

    if (pps == NULL) {
	/* first in list */
	processTable.plist = ps->next;
    } else {
	pps->next = ps->next;
    }

    free(ps);
    processTable.pcnt--;
}

#ifdef HAVE_WAITPID
static void 
cleanChildren()
{
    int                 options;
    int			stat_loc;
    int                 child = 1;

    options = WNOHANG;
    while (child > 0) {
	child = waitpid(-1, &stat_loc, options);
	if (child == 0) {
	    /* child not yet finished so don't complain */
	} else if (child < 0) {
	    if (errno == ECHILD) {
		/* no children so don't complain */
	    } else {
		errmsg("wait for child failed: %s", strerror(errno));
	    }
	} else if (child > 0) {
	    if (verbose) {
		printf("Cleaned up after child (%d)\n", child);
	    }
	/* Remove Entry from Process Table */
	    removeProcessFromTable(child);
	}
    }
}
#elif HAVE_WAIT3
static void 
cleanChildren()
{
#if defined(__NeXT__)
    /* some systems need a union wait as argument to wait3 */
    union wait          status;
#else
    int                 status;
#endif
    int                 options;
    struct rusage       rusage;
    int                 child = 1;

    options = WNOHANG;
    while (child > 0) {
	child = wait3(&status, options, &rusage);
	if (child < 0) {
	    if (errno == ECHILD) {
		/* no children so don't complain */
	    } else {
		errmsg("wait for child failed: %s", strerror(errno));
	    }
	} else if (child > 0) {
	    if (verbose) {
		printf("Cleaned up after child (%d)\n", child);
	    }
	/* Remove Entry from Process Table */
	    removeProcessFromTable(child);
	}
    }
}
#else
/* don't know how to wait for my children */
static void 
cleanChildren()
{
    errmsg("cannot wait for child processes");
}
#endif


/*
 * Refuse an Association
 */

static CONDITION
refuseAssociation(T_ASC_Association ** assoc, CTN_RefuseReason reason)
{
    CONDITION           cond;
    T_ASC_RejectParameters rej;

    if (verbose) {
	printf("Refusing Association (");
	switch (reason) {
	case CTN_TooManyAssociations:
	    printf("TooManyAssociations");
	    break;
	case CTN_CannotFork:
	    printf("CannotFork");
	    break;
	case CTN_BadAppContext:
	    printf("BadAppContext");
	    break;
	case CTN_BadAEPeer:
	    printf("BadAEPeer");
	    break;
	case CTN_BadAEService:
	    printf("BadAEService");
	    break;
	case CTN_NoReason:
	    printf("NoReason");
	    break;
	default:
	    printf("???");
	    break;
	}
	printf(")\n");
    }
    switch (reason) {
    case CTN_TooManyAssociations:
	rej.result = ASC_RESULT_REJECTEDTRANSIENT;
	rej.source = ASC_SOURCE_SERVICEPROVIDER_PRESENTATION_RELATED;
	rej.reason = ASC_REASON_SP_PRES_LOCALLIMITEXCEEDED;
	break;
    case CTN_CannotFork:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEPROVIDER_PRESENTATION_RELATED;
	rej.reason = ASC_REASON_SP_PRES_TEMPORARYCONGESTION;
	break;
    case CTN_BadAppContext:
	rej.result = ASC_RESULT_REJECTEDTRANSIENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_APPCONTEXTNAMENOTSUPPORTED;
	break;
    case CTN_BadAEPeer:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_CALLINGAETITLENOTRECOGNIZED;
	break;
    case CTN_BadAEService:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_CALLEDAETITLENOTRECOGNIZED;
	break;
    case CTN_NoReason:
    default:
	rej.result = ASC_RESULT_REJECTEDPERMANENT;
	rej.source = ASC_SOURCE_SERVICEUSER;
	rej.reason = ASC_REASON_SU_NOREASON;
	break;
    }

    cond = ASC_rejectAssociation(*assoc, &rej);
    if (cond != ASC_NORMAL) {
	fprintf(stderr, "Association Reject Failed:\n");
	COND_DumpConditions();
    }

    cond = ASC_dropAssociation(*assoc);
    if (!SUCCESS(cond)) {
	fprintf(stderr, "Cannot Drop Association:\n");
	COND_DumpConditions();
    }
    cond = ASC_destroyAssociation(assoc);
    if (!SUCCESS(cond)) {
	fprintf(stderr, "Cannot Destroy Association:\n");
	COND_DumpConditions();
    }

    return cond;
}

/*
** Return true if path is a directory 
*/

static BOOL
isaDirectory(const char* path)
{
    BOOL ok = FALSE;
    DIR *d = NULL;
    if ((d = opendir(path)) != NULL) {
	ok = TRUE;
	closedir(d);
    }
    return ok;
}

/*
 * Wait for and negotiate incoming association.
 */

static CONDITION
waitForAssociation(T_ASC_Network * net)
{
    CONDITION           cond;
    int                 pid;
    T_ASC_Association  *assoc;
    char                buf[BUFSIZ];
    int timeout;

    if (singleProcess) {
	timeout = 1000;
    } else {
	if (countChildProcesses() > 0) {
	    timeout = 1;
	} else {
	    timeout = 1000;
	}
    }
    if (ASC_associationWaiting(net, timeout)) {
        cond = ASC_receiveAssociation(net, &assoc, maxReceivePDULength);
        if (!SUCCESS(cond)) {
	    if (verbose) {
	        errmsg("Failed to receive association:");
	        COND_DumpConditions();
	    }
	    goto cleanup;
        }
    } else {
	return ASC_NORMAL;
    }

    if (verbose) {
	printf("Association Received (%s:%s -> %s)\n",
	       assoc->params->DULparams.callingPresentationAddress,
	       assoc->params->DULparams.callingAPTitle,
	       assoc->params->DULparams.calledAPTitle);
    }
    if (debug) {
	printf("Parameters:\n");
	ASC_dumpParameters(assoc->params);
    }
    if (refuse) {
	if (verbose) {
	    printf("Refusing Association (forced via command line)\n");
	}
	cond = refuseAssociation(&assoc, CTN_NoReason);
	goto cleanup;
    }

    /* Application Context Name */
    if ((cond = ASC_getApplicationContextName(assoc->params, buf) !=
	 ASC_NORMAL) || strcmp(buf, DICOM_STDAPPLICATIONCONTEXT) != 0) {
        /* reject: the application context name is not supported */
	if (verbose) {
	    errmsg("Bad AppContextName: %s", buf);
	}
	cond = refuseAssociation(&assoc, CTN_BadAppContext);
	goto cleanup;
    }

    /* Implementation Class UID */
    if (rejectWhenNoImplementationClassUID &&
	strlen(assoc->params->theirImplementationClassUID) == 0) {
        /* reject: no implementation Class UID provided */
	if (verbose) {
	    errmsg("No implementation Class UID provided");
	}
	cond = refuseAssociation(&assoc, CTN_NoReason);
	goto cleanup;
    }


    /* too many concurrent associations ?? 
     */
    if (countChildProcesses() >= maxAssociations) {
	cond = refuseAssociation(&assoc, CTN_TooManyAssociations);
	goto cleanup;
    }

    /* is the called AE title known (which worklist to query)? */
    {
	char dbPath[2048];
	sprintf(dbPath, "%s/%s", baseWorklistDBPath, 
		assoc->params->DULparams.calledAPTitle);
	if (!isaDirectory(dbPath)) {
	    /* there is no subdirectory for this AE Title */
	    cond = refuseAssociation(&assoc, CTN_BadAEService);
	    goto cleanup;
	}
    }

    cond = negotiateAssociation(assoc);
    if (cond != ASC_NORMAL) {
	goto cleanup;
    }

    cond = ASC_acknowledgeAssociation(assoc);
    if (cond != ASC_NORMAL) {
	COND_DumpConditions();
	goto cleanup;
    }
    if (verbose) {
	printf("Association Acknowledged (Max Send PDV: %lu)\n",
	       assoc->sendPDVLength);
	if (ASC_countAcceptedPresentationContexts(assoc->params) == 0)
	    printf("    (but no valid presentation contexts)\n");
	if (debug)
	    ASC_dumpParameters(assoc->params);
    }


    if (singleProcess) {
        /* don't spawn a sub-process to handle the association */
	cond = SCE_handleAssociation(assoc);
    } else {
        /* spawn a sub-process to handle the association */
	pid = fork();
	if (pid < 0) {
	    errmsg("Cannot create association sub-process: %s",
		   strerror(errno));
	    cond = refuseAssociation(&assoc, CTN_CannotFork);
	    goto cleanup;
	} else if (pid > 0) {
	    /* parent process, note process in table */

	    addProcessToTable(pid, assoc);

	} else {
	    /* child process, handle the association */
	    cond = SCE_handleAssociation(assoc);
	    /* the child process is done so exit */
	    exit(0);
	}

    }

cleanup:
    if (!singleProcess) {
        /* the child will handle the association, we can drop it */
        cond = ASC_dropAssociation(assoc);
        if (!SUCCESS(cond)) {
	   errmsg("Cannot Drop Association:");
	   COND_DumpConditions();
        }
        cond = ASC_destroyAssociation(&assoc);
        if (!SUCCESS(cond)) {
	    errmsg("Cannot Destroy Association:");
	    COND_DumpConditions();
        }
    }

    if (verbose) {
        /* are there any conditions sitting on the condition stack? */
	char                buf[BUFSIZ];
	CONDITION           c;

	if (COND_TopCondition(&c, buf, BUFSIZ) != COND_NORMAL) {
	    printf("CONDITIONS Remaining:");
	    COND_DumpConditions();
	}
    }
    COND_PopCondition(TRUE);	/* clean out condition stack */
    return cond;
}


static CONDITION
negotiateAssociation(T_ASC_Association * assoc)
{
    CONDITION cond = ASC_NORMAL;
    DIC_AE calledAETitle;
    const char* transferSyntaxes[] = { 
	NULL, NULL, UID_LittleEndianImplicitTransferSyntax };

    ASC_getAPTitles(assoc->params, NULL, calledAETitle, NULL);

    /* 
    ** We prefer to accept Explicitly encoded transfer syntaxes.
    ** If we are running on a Little Endian machine we prefer 
    ** LittleEndianExplicitTransferSyntax to BigEndianTransferSyntax.
    */

    /* gLocalByteOrder is defined in dcxfer.h */
    if (gLocalByteOrder == EBO_LittleEndian) {
	/* we are on a little endian machine */
	transferSyntaxes[0] = UID_LittleEndianExplicitTransferSyntax;
	transferSyntaxes[1] = UID_BigEndianExplicitTransferSyntax;
    } else {
	/* we are on a big endian machine */
	transferSyntaxes[0] = UID_BigEndianExplicitTransferSyntax;
	transferSyntaxes[1] = UID_LittleEndianExplicitTransferSyntax;
    }

    /*  accept any of the supported abstract syntaxes */
    cond = ASC_acceptContextsWithPreferredTransferSyntaxes(
	assoc->params, 
	(const char**)supportedAbstractSyntaxes, numberOfSupportedAbstractSyntaxes,
	(const char**)transferSyntaxes, DIM_OF(transferSyntaxes));
    if (!SUCCESS(cond)) {
	errmsg("Cannot accept presentation contexts:");
	COND_DumpConditions();
    }

    return cond;
}


/*
** CVS Log
** wlistctn.cc,v
** Revision 1.4  1996/08/29 16:16:10  hewett
** Corrected copyright header.
**
** Revision 1.3  1996/05/31 13:50:44  hewett
** Now checks that the called AE Title corresponds to a Worklist database
** directory.  If no such directory exists then the association is rejected
** due to an unknown AE Title.
**
** Revision 1.2  1996/05/01 17:33:50  hewett
** Modified for latest Worklist DB.
**
** Revision 1.1.1.1  1996/04/26 14:16:01  hewett
** Initial Import.
**
**
*/
